#!/usr/bin/env python3

import sys
from os.path import splitext
import argparse

from AaronTools.geometry import Geometry
from AaronTools.fileIO import FileReader, read_types
from AaronTools.theory import *
from AaronTools.utils.utils import combine_dicts, get_filename, glob_files

theory_parser = argparse.ArgumentParser(
    description="print Gaussian, ORCA, Psi4, SQM, or QChem input file",
    formatter_class=argparse.RawTextHelpFormatter
)

theory_parser.add_argument(
    "infile", metavar="input file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="a coordinate file",
)

theory_parser.add_argument(
    "-o", "--output",
    type=str,
    default=False,
    required=False,
    dest="outfile",
    help="output destination\n" +
    "$INFILE will be replaced with the name of the input file\n" +
    "Default: stdout"
)

theory_parser.add_argument(
    "-if", "--input-format",
    type=str,
    default=None,
    dest="input_format",
    choices=read_types,
    help="file format of input - xyz is assumed if input is stdin",
)

theory_parser.add_argument(
    "-of", "--output-format",
    type=str,
    default=None,
    dest="out_format",
    choices=["gaussian", "orca", "psi4", "sqm", "qchem"],
    help="file format of output",
)

theory_parser.add_argument(
    "-c", "--comment",
    action="append",
    nargs="+",
    default=[],
    dest="comments",
    help="comment to put in the output file\ninput file(s) should not be right after comments",
)

theory_parser.add_argument(
    "-q", "--charge",
    type=int,
    dest="charge",
    default=None,
    help="net charge\nDefault: 0 or what is found in the input file",
)

theory_parser.add_argument(
    "-mult", "--multiplicity",
    type=int,
    dest="multiplicity",
    default=None,
    help="electronic multiplicity\nDefault: 1 or what is found in the input file",
)

theory_parser.add_argument(
    "-p", "--cores",
    type=int,
    dest="processors",
    default=None,
    required=False,
    help="number of cpu cores to use",
)

theory_parser.add_argument(
    "-mem", "--memory",
    type=int,
    dest="memory",
    default=None,
    required=False,
    help="total memory in GB\n" +
    "Note: ORCA and Gaussian only use this to limit the storage-intensive\n" +
    "      portions of the calculation (e.g. integrals, wavefunction info)",
)

theory_parser.add_argument(
    "-up", "--use-previous",
    action="store_true",
    default=False,
    required=False,
    dest="use_prev",
    help="use settings that can be parsed from the input file",
)

theory_options = theory_parser.add_argument_group("Theory options")
theory_options.add_argument(
    "-m", "--method",
    type=str,
    dest="method",
    required=False,
    help="method (e.g. B3LYP or MP2)",
)

theory_options.add_argument(
    "-b", "--basis",
    nargs="+",
    type=str,
    action="append",
    required=False,
    default=None,
    dest="basis",
    help="basis set or list of elements and basis set (e.g. C O N aug-cc-pvtz)\n" +
    "elements can be prefixed with ! to exclude them from the basis\n" +
    "tm is a synonym for d-block elements\n" +
    "auxilliary basis sets can be specified by putting aux X before the basis\n" +
    "set name, where X is the auxilliary type (e.g. aux JK cc-pVDZ for cc-pVDZ/JK)\n" +
    "a path to a file containing a basis set definition (like one\n" +
    "downloaded from basissetexchange.org) can be placed after the\n" +
    "basis set name\n" +
    "the file's contents should be appropriate for the software package you are using"
)

theory_options.add_argument(
    "-ecp", "--pseudopotential",
    nargs="+",
    type=str,
    action="append",
    required=False,
    default=None,
    dest="ecp",
    help="ECP or list of elements and ECP (e.g. Pt LANL2DZ)\n" +
    "elements can be prefixed with ! to exclude them from the ECP\n" +
    "tm is a synonym for d-block elements\n" +
    "a path to a file containing a basis set definition (like one\n" +
    "downloaded from basissetexchange.org) can be placed after the\n" +
    "basis set name\n" +
    "the file's contents should be appropriate for the software package you are using"
)

theory_options.add_argument(
    "-ed", "--empirical-dispersion",
    required=False,
    default=None,
    dest="empirical_dispersion",
    help="empirical dispersion keyword",
)

theory_options.add_argument(
    "-sv", "--solvent",
    required=False,
    default=None,
    dest="solvent",
    help="solvent",
)

theory_options.add_argument(
    "-sm", "--solvent-model",
    required=False,
    default=None,
    dest="solvent_model",
    help="implicit solvent model",
)

theory_options.add_argument(
    "-g", "--grid",
    required=False,
    default=None,
    dest="grid",
    help="integration grid",
)

job_options = theory_parser.add_argument_group("Job options")
job_options.add_argument(
    "-opt", "--optimize",
    action="store_true",
    required=False,
    dest="optimize",
    help="request geometry optimization job",
)

job_options.add_argument(
    "-freq", "--frequencies",
    action="store_true",
    required=False,
    default=False,
    dest="freq",
    help="request vibrational frequencies job",
)

job_options.add_argument(
    "-e", "--energy",
    action="store_true",
    required=False,
    default=False,
    dest="energy",
    help="request single point energy job",
)

opt_type = theory_parser.add_argument_group("Optimization options")
opt_type.add_argument(
    "-ts", "--transition-state",
    action="store_true",
    default=False,
    dest="ts",
    help="request transition state optimization",
)

opt_type.add_argument(
    "-ca", "--constrained-atoms",
    nargs=1,
    type=str,
    action="append",
    default=None,
    dest="atoms",
    help="comma- or hyphen-separated list of atoms (1-indexed) to constrain during optimization",
)

opt_type.add_argument(
    "-cb", "--constrain-bond",
    nargs=1,
    action="append",
    default=None,
    dest="bonds",
    help="list of comma-separated atom pairs\n" +
    "the distance between the atoms in each pair will be constrained during optimization",
)

opt_type.add_argument(
    "-cang", "--constrain-angle",
    type=str,
    nargs=1,
    action="append",
    default=None,
    dest="angles",
    help="list of comma-separated atom trios\n" +
    "the angle defined by each trio will be constrained during optimization",
)

opt_type.add_argument(
    "-ct", "--constrain-torsion",
    type=str,
    nargs=1,
    action="append",
    default=None,
    dest="torsions",
    help="list of comma-separated atom quartets\n" +
    "the torsional angle defined by each quartet will be constrained during optimization",
)

opt_type.add_argument(
    "-cx", "--constrained-x",
    nargs=1,
    type=str,
    action="append",
    default=None,
    dest="x",
    help="comma- or hyphen-separated list of atoms (1-indexed) to constrain the x coordinate of\n" +
    "available for Gaussian, Psi4, and Q-Chem",
)

opt_type.add_argument(
    "-cy", "--constrained-y",
    nargs=1,
    type=str,
    action="append",
    default=None,
    dest="y",
    help="comma- or hyphen-separated list of atoms (1-indexed) to constrain the y coordinate of\n" +
    "available for Gaussian, Psi4, and Q-Chem",
)

opt_type.add_argument(
    "-cz", "--constrained-z",
    nargs=1,
    type=str,
    action="append",
    default=None,
    dest="z",
    help="comma- or hyphen-separated list of atoms (1-indexed) to constrain the z coordinate of\n" +
    "available for Gaussian, Psi4, and Q-Chem",
)

opt_type.add_argument(
    "-gx", "--grouped-x",
    nargs=2,
    type=str,
    action="append",
    default=None,
    dest="xgroup",
    metavar=("atoms", "value"),
    help="comma- or hyphen-separated list of atoms (1-indexed) to keep in the same yz plane\n" +
    "available for Gaussian and Psi4",
)

opt_type.add_argument(
    "-gy", "--grouped-y",
    nargs=2,
    type=str,
    action="append",
    default=None,
    dest="ygroup",
    metavar=("atoms", "value"),
    help="comma- or hyphen-separated list of atoms (1-indexed) to keep in the same xz plane\n" +
    "available for Gaussian, Psi4",
)

opt_type.add_argument(
    "-gz", "--grouped-z",
    nargs=2,
    type=str,
    action="append",
    default=None,
    dest="zgroup",
    metavar=("atoms", "value"),
    help="comma- or hyphen-separated list of atoms (1-indexed) to keep in the same xy plane\n" +
    "available for Gaussian and Psi4",
)

freq_type = theory_parser.add_argument_group("Frequency options")
freq_type.add_argument(
    "-n", "--numerical",
    action="store_true",
    default=False,
    dest="numerical",
    help="request numerical frequencies",
)

freq_type.add_argument(
    "-t", "--temperature",
    type=float,
    default=298.15,
    dest="temperature",
    help="temperature for calculated thermochemical corrections\nDefault: 298.15",
)

qchem_options = theory_parser.add_argument_group("Q-Chem-specific options")
qchem_options.add_argument(
    "--rem",
    action="append",
    nargs="+",
    default=[],
    dest=QCHEM_REM,
    metavar=("KEYWORD", "OPTION"),
    help="REM options\nexample: --rem MAX_SCF_CYCLES 300\n" +
    "input file(s) should not be right after --rem",
)

qchem_options.add_argument(
    "--section",
    nargs="+",
    action="append",
    default=[],
    dest=QCHEM_SETTINGS,
    metavar=("SECTION_NAME", "TEXT"),
    help="add text to a section"
    "input file(s) should not be right after --section",
)


orca_options = theory_parser.add_argument_group("ORCA-specific options")
orca_options.add_argument(
    "--simple",
    action="append",
    default=[],
    dest=ORCA_ROUTE,
    help="keywords for simple input",
)

orca_options.add_argument(
    "--block",
    nargs=3,
    action="append",
    default=[],
    dest=ORCA_BLOCKS,
    metavar=("BLOCK", "OPTION", "VALUE"),
    help="blocks and block options\nexample: --block scf maxiter 500",
)


psi4_options = theory_parser.add_argument_group("Psi4-specific options")
psi4_options.add_argument(
    "--before-molecule",
    action="append",
    default=[],
    dest=PSI4_BEFORE_GEOM,
    metavar="BEFORE MOL",
    help="line to add before the molecule specification",
)

psi4_options.add_argument(
    "--before-job",
    action="append",
    nargs="+",
    default=[],
    dest=PSI4_BEFORE_JOB,
    metavar="BEFORE JOB",
    help="line to add before the job\ninput file(s) should not be right after --before-job",
)

psi4_options.add_argument(
    "--after-job",
    action="append",
    nargs="+",
    default=[],
    dest=PSI4_AFTER_JOB,
    metavar="AFTER JOB",
    help="line to add after the job\ninput file(s) should not be right after --after-job",
)


psi4_options.add_argument(
    "--job",
    action="append",
    nargs="+",
    default=[],
    dest=PSI4_JOB,
    metavar="JOB",
    help="other jobs to add\nexample: --job hessian\ninput file(s) should not be right after --job",
)

psi4_options.add_argument(
    "--setting",
    action="append",
    nargs=2,
    default=[],
    dest=PSI4_SETTINGS,
    metavar=("SETTING", "VALUE"),
    help="settings\nexample: --setting reference uhf",
)

psi4_options.add_argument(
    "--pcm-solver",
    action="append",
    nargs="+",
    default=[],
    dest=PSI4_SOLVENT,
    metavar=("SETTING", "VALUE"),
    help="settings\nexample: --pcm-solver Cavity 'RadiiSet = UFF' 'Area = 0.3'" +
    "\ninput file(s) should not be right after --pcm-solver",
)

psi4_options.add_argument(
    "--optking",
    action="append",
    nargs=2,
    default=[],
    dest=PSI4_OPTKING,
    metavar=("SETTING", "VALUE"),
    help="optking settings",
)

psi4_options.add_argument(
    "--molecule",
    action="append",
    nargs="+",
    default=[],
    dest=PSI4_MOLECULE,
    metavar=("SETTING", "VALUE"),
    help="options to add to the molecule section\n" +
    "example: --molecule units bohr\ninput file(s) should not be right after --molecule",
)


gaussian_options = theory_parser.add_argument_group("Gaussian-specific options")
gaussian_options.add_argument(
    "--route",
    action="append",
    nargs="+",
    default=[],
    dest=GAUSSIAN_ROUTE,
    metavar=("KEYWORD", "OPTION"),
    help="route options\nexample: --route freq hpmodes\n" +
    "input file(s) should not be right after --route",
)

gaussian_options.add_argument(
    "--link0",
    action="append",
    nargs="+",
    default=[],
    dest=GAUSSIAN_PRE_ROUTE,
    metavar=("COMMAND", "VALUE"),
    help="Link 0 commands (without %%)\n" +
    "example: --link0 chk asdf.chk\ninput file(s) should not be right after --link0",
)

gaussian_options.add_argument(
    "--end-of-file",
    action="append",
    default=[],
    dest=GAUSSIAN_POST,
    metavar="input",
    help="line to add to the end of the file (e.g. for NBORead)",
)


args = theory_parser.parse_args()

if not args.method and not args.use_prev:
    sys.stderr.write("no method specified; -m/--method or -u/--use-previous is required")
    theory_parser.print_help()
    sys.exit(1)

kwargs = {}

blocks = getattr(args, ORCA_BLOCKS)
if blocks:
    kwargs[ORCA_BLOCKS] = {}
    for block in blocks:
        block_name = block[0]
        if block_name not in kwargs[ORCA_BLOCKS]:
            kwargs[ORCA_BLOCKS][block_name] = []
        kwargs[ORCA_BLOCKS][block_name].append("\t".join(block[1:]))

for pos in [
        PSI4_SETTINGS, PSI4_MOLECULE, PSI4_JOB, PSI4_OPTKING, PSI4_SOLVENT,
        GAUSSIAN_ROUTE, GAUSSIAN_PRE_ROUTE, QCHEM_REM, QCHEM_SETTINGS,
    ]:
    opts = getattr(args, pos)
    if opts:
        if pos not in kwargs:
            kwargs[pos] = {}

        for opt in opts:
            setting = opt.pop(0)
            if setting.lower() == "rem" and pos == QCHEM_SETTINGS:
                s = " ".join(["'%s'" % val if " " in val else val for val in opt])
                raise TypeError("use --rem %s instead of --section rem %s" % (s, s))
            if setting not in kwargs[pos]:
                kwargs[pos][setting] = []

            kwargs[pos][setting].extend(opt)


for pos in [ORCA_ROUTE]:
    opt = getattr(args, pos)
    if opt:
        if pos not in kwargs:
            kwargs[pos] = []

        kwargs[pos].extend(opt)

for pos in [PSI4_BEFORE_GEOM, PSI4_AFTER_JOB, PSI4_BEFORE_JOB, GAUSSIAN_POST]:
    opt = getattr(args, pos)
    if opt:
        if pos not in kwargs:
            kwargs[pos] = []

        kwargs[pos].extend([" ".join(word) for word in opt])

if args.comments:
    kwargs["comments"] = [" ".join(comment) for comment in args.comments]


# Theory() is made for each file because we might be using things from the input file
for f in glob_files(args.infile, parser=theory_parser):
    if isinstance(f, str):
        if args.input_format is not None:
            infile = FileReader((f, args.input_format[0], None), just_geom=False, get_all=True)
        else:
            infile = FileReader(f, just_geom=False, get_all=True)
    else:
        if args.input_format is not None:
            infile = FileReader(
                ("from stdin", args.input_format[0], f), just_geom=False, get_all=True
            )
        else:
            if len(sys.argv) >= 1:
                infile = FileReader(("from stdin", "xyz", f), just_geom=False, get_all=True)

    geom = Geometry(infile)


    if args.method is None and args.use_prev:
        if "method" in infile.other:
            method = infile.other["method"].split("/")[0]
        elif "theory" in infile.other:
            method = infile.other["theory"].method
    elif args.method is not None:
        method = args.method
    else:
        raise RuntimeError(
            "method was not determined from %s and was not specified with --method" % f
        )


    if args.basis is not None:
        basis_sets = []
        for basis in args.basis:
            basis_sets.append(
                BasisSet.parse_basis_str(" ".join(basis))[0]
            )

    elif args.use_prev:
        if "method" in infile.other:
            basis_sets = method.split("/")[-1]
        elif "theory" in infile.other:
            basis_sets = infile.other["theory"].basis.basis
        else:
            basis_sets = None

    else:
        basis_sets = None


    if args.ecp is not None:
        ecps = []
        for ecp in args.ecp:
            ecps.append(
                BasisSet.parse_basis_str(" ".join(ecp), cls=ECP)[0]
            )

    elif args.use_prev:
        if "theory" in infile.other:
            ecps = infile.other["theory"].basis.ecp
        else:
            ecps = None

    else:
        ecps = None


    if ecps is None and basis_sets is None:
        basis_set = None
    else:
        basis_set = BasisSet(basis_sets, ecps)


    if args.solvent is not None or args.solvent_model is not None:
        if args.solvent_model is None or args.solvent is None:
            raise RuntimeError("--solvent and --solvent-model must both be specified")

        solvent = ImplicitSolvent(args.solvent_model, args.solvent)

    else:
        solvent = None


    job_types = []
    if not args.use_prev or (args.optimize or args.freq or args.energy):
        if args.optimize:
            constraints = {}
            if args.atoms is not None:
                constraints["atoms"] = []
                for constraint in args.atoms:
                    constraints["atoms"].extend(geom.find(constraint))

            if args.bonds is not None:
                constraints["bonds"] = []
                for bond in args.bonds:
                    bonded_atoms = geom.find(bond)
                    if len(bonded_atoms) != 2:
                        raise RuntimeError(
                            "not exactly 2 atoms specified in a bond constraint\n" +
                            "use the format --constrain-bond 1,2"
                        )
                    constraints["bonds"].append(bonded_atoms)

            if args.angles is not None:
                constraints["angles"] = []
                for angle in args.angle:
                    angle_atoms = geom.find(angle)
                    if len(angle_atoms) != 3:
                        raise RuntimeError(
                            "not exactly 3 atoms specified in a angle constraint\n" +
                            "use the format --constrain-angle 1,2,3"
                        )
                    constraints["angles"].append(angle_atoms)

            if args.torsions is not None:
                constraints["torsions"] = []
                for torsion in args.torsions:
                    torsion_atoms = geom.find(torsion)
                    if len(torsion_atoms) != 4:
                        raise RuntimeError(
                            "not exactly 4 atoms specified in a torsion constraint\n" +
                            "use the format --constrain-torsion 1,2,3,4"
                        )
                    constraints["torsions"].append(torsion_atoms)

            if args.x is not None:
                constraints["x"] = []
                for constraint in args.x:
                    constraints["x"].extend(geom.find(constraint))

            if args.y is not None:
                constraints["y"] = []
                for constraint in args.y:
                    constraints["y"].extend(geom.find(constraint))

            if args.z is not None:
                constraints["z"] = []
                for constraint in args.z:
                    constraints["z"].extend(geom.find(constraint))

            if args.xgroup is not None:
                constraints["xgroup"] = []
                for constraint, val in args.xgroup:
                    constraints["xgroup"].append((geom.find(constraint), float(val)))

            if args.ygroup is not None:
                constraints["ygroup"] = []
                for constraint, val in args.ygroup:
                    constraints["ygroup"].append((geom.find(constraint), float(val)))

            if args.zgroup is not None:
                constraints["zgroup"] = []
                for constraint, val in args.zgroup:
                    constraints["zgroup"].append((geom.find(constraint), float(val)))

            if not constraints.keys():
                constraints = None

            job_types.append(OptimizationJob(transition_state=args.ts, constraints=constraints))


        if args.freq:
            job_types.append(FrequencyJob(numerical=args.numerical, temperature=args.temperature))

        if args.energy:
            job_types.append(SinglePointJob())

    elif args.use_prev and "theory" in infile.other:
        job_types = infile.other["theory"].job_type

    grid = args.grid
    if args.use_prev and "theory" in infile.other and not grid:
        grid = infile.other["theory"].grid

    if args.charge is None:
        if "charge" in infile.other:
            charge = infile.other["charge"]
        else:
            charge = 0
    else:
        charge = args.charge


    if args.multiplicity is None:
        if "multiplicity" in infile.other:
            multiplicity = infile.other["multiplicity"]
        else:
            multiplicity = 1
    else:
        multiplicity = args.multiplicity


    other_kwargs = {}

    theory = Theory(
        method=method,
        basis=basis_set,
        grid=grid,
        solvent=solvent,
        job_type=job_types,
        empirical_dispersion=args.empirical_dispersion,
        charge=charge,
        multiplicity=multiplicity,
        processors=args.processors,
        memory=args.memory,
    )


    if args.out_format:
        style = args.out_format
    else:
        if args.outfile:
            style = splitext(args.outfile)[-1].lstrip(".")
        else:
            raise RuntimeError("file format must be specified if no output file is specified")

    if args.use_prev and "other_kwargs" in infile.other:
        other_kwargs = combine_dicts(other_kwargs, infile.other["other_kwargs"])

    other_kwargs = combine_dicts(kwargs, other_kwargs)

    if args.outfile:
        outfile = args.outfile
        if "$INFILE" in outfile:
            outfile = outfile.replace("$INFILE", get_filename(f))
        warnings = geom.write(
            append=True,
            outfile=outfile,
            style=style,
            theory=theory,
            return_warnings=True,
            **other_kwargs
        )
    else:
        out, warnings = geom.write(
            append=True,
            outfile=False,
            style=style,
            theory=theory,
            return_warnings=True,
            **other_kwargs
        )
        print(out)
    
    if warnings:
        for warning in warnings:
            Geometry.LOG.warning(warning)



#!/usr/bin/env python3

import argparse
import os

from AaronTools.geometry import Geometry
from AaronTools.component import Component

vsepr_choices = [
        "tetrahedral",
        "seesaw",
        "square_planar",
        "trigonal_pyramidal",
        "trigonal_bipyramidal",
        "square_pyramidal",
        "pentagonal",
        "hexagonal",
        "trigonal_prismatic",
        "pentagonal_pyramidal",
        "octahedral",
        "capped_octahedral",
        "hexagonal_pyramidal",
        "pentagonal_bipyramidal",
        "capped_trigonal_prismatic",
        "heptagonal",
]

coord_comp_parser = argparse.ArgumentParser(
    description="build coordination complexes using templates from Inorg. Chem. 2018, 57, 17, 10557–10567",
    formatter_class=argparse.RawTextHelpFormatter,
)

coord_comp_parser.add_argument(
    "-l",
    "--ligands",
    type=str,
    nargs="*",
    required=True,
    # choices=Component.list(),
    dest="ligands",
    help="list of ligands to attach to the coordination complex\n" +
    "see `mapLigand.py --list` for a list of available ligands",
)

coord_comp_parser.add_argument(
    "-c2",
    "--c2-symmetric",
    type=lambda x: x.lower() in ["yes", "true", "t", "arr", "y", "aye", "yeah"],
    nargs="*",
    default=None,
    required=False,
    dest="c2_symmetric",
    help="list of true/false corresping to --ligands to denote which bidentate\n" +
    "ligands are C2-symmetric\nDefault: try to determine if bidentate ligands are C2-symmetric",
)

coord_comp_parser.add_argument(
    "-g",
    "--coordination-geometry",
    choices=vsepr_choices,
    required=True,
    dest="shape",
    help="coordination geometry of central atom"
)
coord_comp_parser.add_argument(
    "-c",
    "--center-atom",
    required=True,
    metavar="element",
    dest="center",
    help="central atom for coordination complexes"
)

coord_comp_parser.add_argument(
    "-m",
    "--minimize",
    action="store_true",
    default=False,
    required=False,
    dest="minimize",
    help="try to relax ligands to minimize steric clashing\nDefault: False",
)

coord_comp_parser.add_argument(
    "-o",
    "--output",
    type=str,
    required=True,
    metavar="output destination",
    dest="outdir",
    help="output directory\n" +\
    "Filenames will match the detected generic formula and\n" +
    "include the point group and subset from the reference\n"
    "noted above\n"
    "Subsets with primes (e.g. A' and A'') are not distinguished",
)

args = coord_comp_parser.parse_args()

geoms, formula = Geometry.get_coordination_complexes(
    args.center,
    args.ligands,
    args.shape.replace("_", " "),
    c2_symmetric=args.c2_symmetric,
    minimize=args.minimize,
)

print("formula is %s" % formula)

if not os.path.exists(args.outdir):
    os.makedirs(args.outdir)

for geom in geoms:
    geom.write(outfile=os.path.join(args.outdir, geom.name + ".xyz"))

print("wrote", len(geoms), "structures to", args.outdir)


"""methods (e.g. DFT functionals, coupled-cluster methods) for Theory()"""
import re

from AaronTools import addlogger

KNOWN_SEMI_EMPIRICAL = [
    "AM1",
    "PM3",
    "PM6",
    "PM7",
    "HF-3C",
    "PM3MM",
    "PDDG",
    "RM1",
    "MNDO",
    "PM3-PDDG",
    "MNDO-PDDG",
    "PM3-CARB1",
    "MNDO/d",
    "AM1/d",
    "DFTB2",
    "DFTB3",
    "AM1-D*",
    "PM6-D",
    "AM1-DH+",
    "PM6-DH+",
]

class Method:
    """functional object
    used to ensure the proper keyword is used
    e.g.
    using Functional('PBE0') will use PBE1PBE in a gaussian input file"""

    LOG = None

    def __init__(self, name, is_semiempirical=False):
        """
        name: str, functional name
        is_semiempirical: bool, basis set is not required
        """
        self.name = name
        self.is_semiempirical = is_semiempirical

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return False
        return (
            self.get_gaussian()[0].lower() == other.get_gaussian()[0].lower() and
            self.is_semiempirical == other.is_semiempirical
        )

    @staticmethod
    def sanity_check_method(name, program):
        """
        check to see if method is available in the specified program
        name - str, name of method
        program, str, gaussian, orca, psi4, or qchem
        """
        import os.path
        from difflib import SequenceMatcher as seqmatch
        from numpy import argsort, loadtxt
        from AaronTools.const import AARONTOOLS
        
        warning = None
        
        prefix = ""
        if program.lower() == "gaussian":
            valid = loadtxt(os.path.join(AARONTOOLS, "theory", "valid_methods", "gaussian.txt"), dtype=str)
            prefix = "(?:RO|R|U)?"
        elif program.lower() == "orca":
            valid = loadtxt(os.path.join(AARONTOOLS, "theory", "valid_methods", "orca.txt"), dtype=str)
        elif program.lower() == "psi4":
            valid = loadtxt(os.path.join(AARONTOOLS, "theory", "valid_methods", "psi4.txt"), dtype=str)
        elif program.lower() == "sqm":
            valid = loadtxt(os.path.join(AARONTOOLS, "theory", "valid_methods", "sqm.txt"), dtype=str)
        elif program.lower() == "qchem":
            valid = loadtxt(os.path.join(AARONTOOLS, "theory", "valid_methods", "qchem.txt"), dtype=str)
        else:
            raise NotImplementedError("cannot validate method names for %s" % program)
        
        if not any(
            # need to escape () b/c they aren't capturing groups, it's ccsd(t) or something
            re.match(
                "%s%s$" % (prefix, method.replace("(", "\(").replace(")", "\)").replace("+", "\+")), name, flags=re.IGNORECASE
            ) for method in valid
        ):
            warning = "method '%s' may not be available in %s\n" % (name, program) + \
            "if this is incorrect, please submit a bug report at https://github.com/QChASM/AaronTools.py/issues"
            
            # try to suggest alternatives that have similar names
            simm = [
                seqmatch(
                    lambda x: x in "-_()/", name.upper(), test_method.upper()
                ).ratio() for test_method in valid
            ]
            ndx = argsort(simm)[-5:][::-1]
            warning += "\npossible misspelling of:\n"
            warning += "\n".join([valid[i] for i in ndx])

        return warning

    def copy(self):
        new_dict = dict()
        for key, value in self.__dict__.items():
            try:
                new_dict[key] = value.copy()
            except AttributeError:
                new_dict[key] = value
                # ignore chimerax objects so seqcrow doesn't print a
                # warning when a geometry is copied
                if "chimerax" in value.__class__.__module__:
                    continue
                if value.__class__.__module__ != "builtins":
                    self.LOG.warning(
                        "No copy method for {}: in-place changes may occur".format(
                            type(value)
                        )
                    )
        
        return self.__class__(**new_dict)

    def get_gaussian(self):
        """maps proper functional name to one Gaussian accepts"""
        warning = None
        if self.name.lower() == "ωb97x-d" or self.name.lower() == "wb97x-d":
            return ("wB97XD", None)
        elif self.name == "Gaussian's B3LYP":
            return ("B3LYP", None)
        elif self.name.lower() == "b97-d":
            return ("B97D", None)
        elif self.name.lower().startswith("m06-"):
            return (self.name.upper().replace("M06-", "M06", 1), None)

        elif self.name.upper() == "PBE0":
            return ("PBE1PBE", None)

        #methods available in ORCA but not Gaussian
        elif self.name.lower() == "ωb97x-d3":
            return ("wB97XD", "ωB97X-D3 is not available in Gaussian, switching to ωB97X-D2")
        elif self.name.lower() == "b3lyp":
            return ("B3LYP", None)
        
        name = self.name.replace('ω', 'w')
        return name, warning

    def get_orca(self):
        """maps proper functional name to one ORCA accepts"""
        warning = None
        if (
                self.name == "ωB97X-D" or
                any(
                    test == self.name.lower() for test in ["wb97xd", "wb97x-d"]
                )
        ):
            return ("wB97X-D3", "ωB97X-D may refer to ωB97X-D2 or ωB97X-D3 - using the latter")
        elif self.name == "ωB97X-D3":
            return ("wB97X-D3", None)
        elif any(self.name.upper() == name for name in ["B97-D", "B97D"]):
            return ("B97-D", None)
        elif self.name == "Gaussian's B3LYP":
            return ("B3LYP/G", None)
        elif self.name.upper() == "M06-L":
            return ("M06L", None)
        elif self.name.upper() == "M06-2X":
            return ("M062X", None)
        elif self.name.upper() == "PBE1PBE":
            return ("PBE0", None)

        name = self.name.replace('ω', 'w')
        return name, warning

    def get_psi4(self):
        """maps proper functional name to one Psi4 accepts"""
        if self.name.lower() == 'wb97xd':
            return "wB97X-D", None
        elif self.name.upper() == 'B97D':
            return ("B97-D", None)
        elif self.name.upper() == "PBE1PBE":
            return ("PBE0", None)
        elif self.name.upper() == "M062X":
            return ("M06-2X", None)
        elif self.name.upper() == "M06L":
            return ("M06-L", None)

        # the functionals havent been combined with dispersion yet, so
        # we aren't checking if the method is available

        return self.name.replace('ω', 'w'), None

    def get_sqm(self):
        """get method name that is appropriate for sqm"""
        return self.name

    def get_qchem(self):
        """maps proper functional name to one Psi4 accepts"""
        if re.match("[wω]b97[xm]?[^-xm]", self.name.lower()) and self.name.lower() != "wb97m(2)":
            name = re.match("([wω]?)b97([xm]?)([\S]+)", self.name.lower())
            return ("%sB97%s%s" % (
                    name.group(1) if name.group(1) else "",
                    name.group(2).upper() if name.group(2) else "",
                    "-%s" % name.group(3).upper() if name.group(3) else "",
                ),
                None
            )
        elif self.name.upper() == 'B97D':
            return ("B97-D", None)
        elif self.name.upper() == "M062X":
            return ("M06-2X", None)
        elif self.name.upper() == "M06L":
            return ("M06-L", None)

        # the functionals havent been combined with dispersion yet, so
        # we aren't checking if the method is available

        return self.name.replace('ω', 'w'), None


class SAPTMethod(Method):
    """
    method used to differentiate between regular methods and sapt
    methods because the molecule will need to be split into monomers
    if using a sapt method, the geometry given to Theory or Geometry.write
    should have a 'components' attribute with each monomer being a coordinate
    the charge and multiplicity given to Theory should be a list, with the first
    item in each list being the overall charge/multiplicity and the subsequent items
    being the charge/multiplicity of the monomers (components)
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


#! /usr/bin/env python3
from collections.abc import MutableSequence

import numpy as np

from AaronTools import addlogger, getlogger
from AaronTools.atoms import Atom
from AaronTools.const import PHYSICAL, UNIT
from AaronTools.fileIO import FileReader
from AaronTools.geometry import Geometry
from AaronTools.utils.utils import float_vec, uptri2sym


def obj_to_dict(obj, skip_attrs=None):
    # log = getlogger(level="debug")
    if skip_attrs is None:
        skip_attrs = []
    if isinstance(obj, Geometry):
        return obj.comment, [str(a) for a in obj]
    rv = {}
    if hasattr(obj, "__dict__"):
        for attr in obj.__dict__:
            if attr in skip_attrs:
                continue
            val = getattr(obj, attr)
            if isinstance(val, MutableSequence):
                val = [obj_to_dict(v) for v in val]
            else:
                val = obj_to_dict(val)
            rv[str(attr)] = val
        return rv
    return obj


@addlogger
class CompOutput:
    """
    Attributes:
        geometry    the last Geometry
        opts        list of Geometry for each optimization steps
        frequency   Frequency object
        archive     a string containing the archive entry
        energy, enthalpy, free_energy, grimme_g,
        mass, temperature, rotational_temperature,
        multiplicity, charge, rotational_symmetry_number
        error, error_msg, finished,
        gradient, E_ZPVE, ZPVE
    """

    ELECTRONIC_ENERGY = "NRG"
    ZEROPOINT_ENERGY = "ZPE"
    RRHO_ENTHALPY = "ENTHALPY"
    QUASI_HARMONIC = "QHARM"
    QUASI_RRHO = "QRRHO"
    RRHO = "RRHO"
    LOG = None
    LOGLEVEL = "debug"

    def __init__(
        self,
        fname="",
        get_all=True,
        freq_name=None,
        conf_name=None,
    ):
        self.geometry = None
        self.opts = None
        self.opt_steps = None
        self.frequency = None
        self.archive = None
        self.other = None
        self.conformers = None

        self.gradient, self.E_ZPVE, self.ZPVE = ({}, None, None)
        self.energy, self.enthalpy = (None, None)
        self.free_energy, self.grimme_g = (None, None)

        self.mass, self.temperature = (None, None)
        self.multiplicity, self.charge = (None, None)

        self.rotational_temperature = None
        self.rotational_symmetry_number = None

        self.error, self.error_msg, self.finished = (None, None, None)

        # these will be pulled out of FileReader.other dict
        keys = [
            "opt_steps",
            "energy",
            "error",
            "error_msg",
            "gradient",
            "finished",
            "frequency",
            "mass",
            "temperature",
            "rotational_temperature",
            "free_energy",
            "multiplicity",
            "charge",
            "E_ZPVE",
            "ZPVE",
            "rotational_symmetry_number",
            "enthalpy",
            "archive",
        ]

        if isinstance(fname, (str, tuple)) and len(fname) > 0:
            from_file = FileReader(
                fname,
                get_all,
                just_geom=False,
                freq_name=freq_name,
                conf_name=conf_name,
            )
        elif isinstance(fname, FileReader):
            from_file = fname
        else:
            return

        if from_file.atoms:
            self.geometry = Geometry(
                from_file.atoms, comment=from_file.comment, name=from_file.name,
            )
        if from_file.all_geom:
            self.opts = []
            for g in from_file.all_geom:
                self.opts += [Geometry(g[0])]
        if "conformers" in from_file.other:
            self.conformers = []
            for comment, atoms in from_file.other["conformers"]:
                self.conformers.append(Geometry(atoms, comment=comment))
            del from_file.other["conformers"]

        for k in keys:
            if k in from_file.other:
                setattr(self, k, from_file.other[k])
            else:
                setattr(self, k, None)
        
        self.other = {k:v for k, v in from_file.other.items() if k not in keys}

        if self.rotational_temperature is None and self.geometry:
            self.compute_rot_temps()

        if self.frequency:
            self.grimme_g = self.calc_Grimme_G()
            # recalculate ZPVE b/c our constants and the ones in various programs
            # might be slightly different
            self.ZPVE = self.calc_zpe()
            self.E_ZPVE = self.energy + self.ZPVE

    @staticmethod
    def boltzmann_weights(
        thermo_cos,
        nrg_cos=None,
        weighting="RRHO",
        temperature=298.15,
        v0=100,
    ):
        """
        returns boltzmann weights
        thermo_cos - list of CompOutput instances for thermochem corrections
        nrg_cos - list of CompOutput to take the electronic energy from
            order should correspond to thermo_cos
            if not given, the energies from thermo_cos are used
        weighting - type of energy to use for weighting
            can be:
                "NRG"
                "ZPE"
                "ENTHALPY"
                "QHARM"
                "QRRHO"
                "RRHO"
        temperature - temperature in K
        v0 - parameter for quasi free energy corrections
        """
        if not nrg_cos:
            nrg_cos = thermo_cos
        energies = np.array([co.energy for co in nrg_cos])
        corr = None
        if weighting == CompOutput.ZEROPOINT_ENERGY:
            corr = np.array([co.ZPVE for co in thermo_cos])
        elif weighting == CompOutput.RRHO_ENTHALPY:
            corr = np.array([
                co.therm_corr(temperature=temperature, v0=v0)[1] for
                co in thermo_cos
            ])
        elif weighting == CompOutput.QUASI_HARMONIC:
            corr = np.array([
                co.calc_G_corr(temperature=temperature, v0=v0, method=weighting) for
                co in thermo_cos
            ])
        elif weighting == CompOutput.QUASI_RRHO:
            corr = np.array([
                co.calc_G_corr(temperature=temperature, v0=v0, method=weighting) for
                co in thermo_cos
            ])
        elif weighting == CompOutput.RRHO:
            corr = np.array([
                co.calc_G_corr(temperature=temperature, v0=v0, method=weighting) for
                co in thermo_cos
            ])
        if corr is not None:
            try:
                energies += corr
            except ValueError:
                raise RuntimeError(
                    "number of single point energies (%i) "
                    "does not match number of thermochemical "
                    "corrections (%i)" % (len(energies), len(corr))
                )
        relative = energies - min(energies)
        w = np.exp(
            -relative * UNIT.HART_TO_KCAL / (PHYSICAL.R * temperature)
        )
        return w / sum(w)

    def to_dict(self, skip_attrs=None):
        return obj_to_dict(self, skip_attrs=skip_attrs)

    def get_progress(self):
        rv = ""
        grad = self.gradient
        if not grad:
            rv += "Progress not found"
            return rv

        for name in grad:
            rv += "{:>9}:{}/{:<3} ".format(
                name,
                grad[name]["value"],
                "YES" if grad[name]["converged"] else "NO",
            )

        return rv.rstrip()

    def calc_zpe(self, anharmonic=False):
        """returns ZPVE correction"""
        hc = PHYSICAL.PLANCK * PHYSICAL.SPEED_OF_LIGHT / UNIT.HART_TO_JOULE
        if anharmonic:
            vib = sum(self.frequency.real_frequencies)
            x = np.tril(self.other["X_matrix"]).sum()
            x0 = self.other["X0"]
            zpve = hc * (0.5 * vib + 0.25 * x + x0)
        else:
            vib = sum(self.frequency.real_frequencies)
            zpve = 0.5 * hc * vib
        return zpve

    def therm_corr(self, temperature=None, v0=100, method="RRHO", pressure=1):
        """
        returns thermal correction to energy, enthalpy correction to energy, and entropy
        for the specified cutoff frequency and temperature
        in that order (Hartrees for corrections, Eh/K for entropy)

        temperature: float, temperature in K- None will use self.temperature
        pressure: float, pressure in atm
        v0: float, cutoff/damping parameter for quasi G corrections
        method: str - type of quasi treatment:
                RRHO  - no quasi treatment
                QRRHO - Grimme's quasi-RRHO
                see Grimme, S. (2012), Supramolecular Binding Thermodynamics by
                Dispersion‐Corrected Density Functional Theory. Chem. Eur. J.,
                18: 9955-9964. (DOI: 10.1002/chem.201200497) for details
                QHARM - Truhlar's quasi-harmonic
                see J. Phys. Chem. B 2011, 115, 49, 14556–14562
                (DOI: 10.1021/jp205508z) for details
        """
        if self.frequency is None:
            msg = "Vibrational frequencies not found, "
            msg += "cannot calculate vibrational entropy."
            raise AttributeError(msg)

        rot = [temp for temp in self.rotational_temperature if temp != 0]
        T = temperature if temperature is not None else self.temperature
        if T == 0:
            return 0, 0, 0

        if pressure is None:
            pressure = PHYSICAL.STANDARD_PRESSURE
        else:
            pressure *= UNIT.ATM_TO_PASCAL

        mass = self.mass
        sigmar = self.rotational_symmetry_number
        if sigmar is None and len(self.geometry.atoms) == 1:
            sigmar = 3
        mult = self.multiplicity
        freqs = np.array(self.frequency.real_frequencies)

        vib_unit_convert = (
            PHYSICAL.SPEED_OF_LIGHT * PHYSICAL.PLANCK / PHYSICAL.KB
        )
        vibtemps = np.array(
            [f_i * vib_unit_convert for f_i in freqs if f_i > 0]
        )
        if method == "QHARM":
            harm_vibtemps = np.array(
                [
                    f_i * vib_unit_convert
                    if f_i > v0
                    else v0 * vib_unit_convert
                    for f_i in freqs
                    if f_i > 0
                ]
            )
        else:
            harm_vibtemps = vibtemps

        Bav = PHYSICAL.PLANCK ** 2 / (24 * np.pi ** 2 * PHYSICAL.KB)
        Bav *= sum([1 / r for r in rot])

        # Translational
        qt = 2 * np.pi * mass * PHYSICAL.KB * T / (PHYSICAL.PLANCK ** 2)
        qt = qt ** (3 / 2)
        qt *= PHYSICAL.KB * T / pressure
        St = PHYSICAL.GAS_CONSTANT * (np.log(qt) + (5 / 2))
        Et = 3 * PHYSICAL.GAS_CONSTANT * T / 2

        # Electronic
        Se = PHYSICAL.GAS_CONSTANT * (np.log(mult))

        # Rotational
        if all(r == np.inf for r in rot):
            # atomic
            qr = 1
            Sr = 0
        elif len(rot) == 3:
            # non linear molecules
            qr = np.sqrt(np.pi) / sigmar
            qr *= T ** (3 / 2) / np.sqrt(rot[0] * rot[1] * rot[2])
            Sr = PHYSICAL.GAS_CONSTANT * (np.log(qr) + 3 / 2)
        elif len(rot) == 2:
            # linear molecules
            qr = (1 / sigmar) * (T / np.sqrt(rot[0] * rot[1]))
            Sr = PHYSICAL.GAS_CONSTANT * (np.log(qr) + 1)
        else:
            # atoms
            qr = 1
            Sr = 0

        if all(r == np.inf for r in rot):
            Er = 0
        else:
            Er = len(rot) * PHYSICAL.GAS_CONSTANT * T / 2

        # Vibrational
        if method == self.QUASI_HARMONIC:
            Sv = np.sum(
                harm_vibtemps / (T * (np.exp(harm_vibtemps / T) - 1))
                - np.log(1 - np.exp(-harm_vibtemps / T))
            )
        elif method == self.RRHO:
            Sv = np.sum(
                vibtemps / (T * (np.exp(vibtemps / T) - 1))
                - np.log(1 - np.exp(-vibtemps / T))
            )
        elif method == self.QUASI_RRHO:
            mu = PHYSICAL.PLANCK
            mu /= 8 * np.pi ** 2 * freqs * PHYSICAL.SPEED_OF_LIGHT
            mu = mu * Bav / (mu + Bav)
            Sr_eff = 1 / 2 + np.log(
                np.sqrt(
                    8
                    * np.pi ** 3
                    * mu
                    * PHYSICAL.KB
                    * T
                    / PHYSICAL.PLANCK ** 2
                )
            )

            weights = weight = 1 / (1 + (v0 / freqs) ** 4)

            Sv = np.sum(
                weights
                * (
                    harm_vibtemps / (T * (np.exp(harm_vibtemps / T) - 1))
                    - np.log(1 - np.exp(-harm_vibtemps / T))
                )
                + (1 - weights) * Sr_eff
            )

        Ev = np.sum(vibtemps * (1.0 / 2 + 1 / (np.exp(vibtemps / T) - 1)))

        Ev *= PHYSICAL.GAS_CONSTANT
        Sv *= PHYSICAL.GAS_CONSTANT

        Ecorr = (Et + Er + Ev) / (UNIT.HART_TO_KCAL * 1000)
        Hcorr = Ecorr + (
            PHYSICAL.GAS_CONSTANT * T / (UNIT.HART_TO_KCAL * 1000)
        )
        Stot = (St + Sr + Sv + Se) / (UNIT.HART_TO_KCAL * 1000)

        return Ecorr, Hcorr, Stot

    def calc_G_corr(self, temperature=None, v0=0, method="RRHO", **kwargs):
        """
        returns quasi rrho free energy correction (Eh)
        temperature: float, temperature; default is self.temperature
        v0: float, parameter for quasi-rrho or quasi-harmonic entropy
        method: str (RRHO, QRRHO, QHARM) method for treating entropy
                see CompOutput.therm_corr for references
        """
        Ecorr, Hcorr, Stot = self.therm_corr(temperature, v0, method, **kwargs)
        T = temperature if temperature is not None else self.temperature
        Gcorr_qRRHO = Hcorr - T * Stot

        return Gcorr_qRRHO

    def calc_Grimme_G(self, temperature=None, v0=100, **kwargs):
        """
        returns quasi rrho free energy (Eh)
        see Grimme, S. (2012), Supramolecular Binding Thermodynamics by
        Dispersion‐Corrected Density Functional Theory. Chem. Eur. J.,
        18: 9955-9964. (DOI: 10.1002/chem.201200497) for details
        """
        Gcorr_qRRHO = self.calc_G_corr(
            temperature=temperature, v0=v0, method=self.QUASI_RRHO, **kwargs
        )
        return Gcorr_qRRHO + self.energy

    def bond_change(self, atom1, atom2, threshold=0.25):
        """"""
        ref = self.opts[0]
        d_ref = ref.atoms[atom1].dist(ref.atoms[atom2])

        n = len(self.opts) - 1
        for i, step in enumerate(self.opts[::-1]):
            d = step.atoms[atom1].dist(step.atoms[atom2])
            if abs(d_ref - d) < threshold:
                n = len(self.opts) - 1 - i
                break
        return n

    def parse_archive(self):
        """
        Reads info from archive string

        Returns: a dictionary with the parsed information
        """

        def grab_coords(line):
            rv = {}
            for i, word in enumerate(line.split("\\")):
                word = word.split(",")
                if i == 0:
                    rv["charge"] = int(word[0])
                    rv["multiplicity"] = int(word[1])
                    rv["atoms"] = []
                    continue
                rv["atoms"] += [
                    Atom(element=word[0], coords=word[1:4], name=str(i))
                ]
            return rv

        rv = {}
        lines = iter(self.archive.split("\\\\"))
        for line in lines:
            line = line.strip()
            if not line:
                continue

            if line.startswith("@"):
                line = line[1:]
                for word in line.split("\\"):
                    if "summary" not in rv:
                        rv["summary"] = [word]
                    elif word not in rv["summary"]:
                        rv["summary"] += [word]
                continue

            if line.startswith("#"):
                if "route" not in rv:
                    rv["route"] = line
                elif isinstance(rv["route"], list):
                    # for compound jobs, like opt freq
                    rv["route"] += [line]
                else:
                    # for compound jobs, like opt freq
                    rv["route"] = [rv["route"]] + [line]

                line = next(lines).strip()
                if "comment" not in line:
                    rv["comment"] = line

                line = next(lines).strip()
                for key, val in grab_coords(line).items():
                    rv[key] = val
                continue

            words = iter(line.split("\\"))
            for word in words:
                if not word:
                    # get rid of pesky empty elements
                    continue
                if "=" in word:
                    key, val = word.split("=")
                    rv[key.lower()] = float_vec(val)
                else:
                    if "hessian" not in rv:
                        rv["hessian"] = uptri2sym(
                            float_vec(word),
                            3 * len(rv["atoms"]),
                            col_based=True,
                        )
                    else:
                        rv["gradient"] = float_vec(word)
        return rv

    def follow(self, reverse=False, step=0.1):
        """
        Follow imaginary mode
        """
        # get geometry and frequency objects
        geom = self.geometry.copy()
        freq = self.frequency

        # make sure geom is a TS and has computed frequencies available
        if freq is None:
            raise AttributeError("Frequencies for this geometry not found.")
        if not freq.is_TS:
            raise RuntimeError("Geometry not a transition state")

        # get displacement vectors for imaginary frequency
        img_mode = freq.imaginary_frequencies[0]
        vector = freq.by_frequency[img_mode]["vector"]

        # apply transformation to geometry and return it
        for i, a in enumerate(geom.atoms):
            if reverse:
                a.coords -= vector[i] * step
            else:
                a.coords += vector[i] * step
        return geom

    def compute_rot_temps(self):
        """
        sets self's 'rotational_temperature' attribute by using self.geometry
        not recommended b/c atoms should be specific isotopes, but this uses
        average atomic weights
        exists because older versions of ORCA don't print rotational temperatures
        """
        COM = self.geometry.COM(mass_weight=True)
        self.geometry.coord_shift(-COM)
        inertia_mat = np.zeros((3, 3))
        for atom in self.geometry.atoms:
            for i in range(0, 3):
                for j in range(0, 3):
                    if i == j:
                        inertia_mat[i][j] += sum(
                            [
                                atom.mass() * atom.coords[k] ** 2
                                for k in range(0, 3)
                                if k != i
                            ]
                        )
                    else:
                        inertia_mat[i][j] -= (
                            atom.mass() * atom.coords[i] * atom.coords[j]
                        )

        principal_inertia, vecs = np.linalg.eigh(inertia_mat)

        principal_inertia *= UNIT.AMU_TO_KG * 1e-20

        # rotational constants in Hz
        rot_consts = [
            PHYSICAL.PLANCK / (8 * np.pi ** 2 * moment)
            for moment in principal_inertia
            if moment > 0
        ]

        self.rotational_temperature = [
            PHYSICAL.PLANCK * const / PHYSICAL.KB for const in rot_consts
        ]
        # shift geometry back
        self.geometry.coord_shift(COM)



#!/usr/bin/env python3

import argparse
import sys

from AaronTools.fileIO import FileReader, read_types
from AaronTools.geometry import Geometry
from AaronTools.substituent import Substituent
from AaronTools.utils.utils import glob_files

def main(argv):
    sterimol_parser = argparse.ArgumentParser(
        description="calculate B1-B5, and L sterimol parameters for substituents - see Verloop, A. and Tipker, J. (1976), Use of linear free energy related and other parameters in the study of fungicidal selectivity. Pestic. Sci., 7: 379-390.",
        formatter_class=argparse.RawTextHelpFormatter
    )
    
    sterimol_parser.add_argument(
        "infile", metavar="input file",
        type=str,
        nargs="*",
        default=[sys.stdin],
        help="a coordinate file"
    )
    
    sterimol_parser.add_argument(
        "-if", "--input-format",
        type=str,
        default=None,
        choices=read_types,
        dest="input_format",
        help="file format of input\nxyz is assumed if input is stdin"
    )
    
    sterimol_parser.add_argument(
        "-s", "--substituent-atom",
        type=str,
        required=True,
        dest="targets",
        help="substituent atom\n" +
        "1-indexed position of the starting position of the\n" +
        "substituent of which you are calculating sterimol\nparameters"
    )
    
    sterimol_parser.add_argument(
        "-a", "--attached-to",
        type=str,
        required=True,
        dest="avoid",
        help="non-substituent atom\n" +
        "1-indexed position of the starting position of the atom\n" +
        "connected to the substituent of which you are calculating\n" +
        "sterimol parameters"
    )
    
    sterimol_parser.add_argument(
        "-r", "--radii",
        type=str,
        default="bondi",
        choices=["bondi", "umn"],
        dest="radii",
        help="VDW radii to use in calculation\n" + 
        "umn: main group vdw radii from J. Phys. Chem. A 2009, 113, 19, 5806–5812\n" +
        "    (DOI: 10.1021/jp8111556)\n" + 
        "    transition metals are crystal radii from Batsanov, S.S. Van der Waals\n" +
        "    Radii of Elements. Inorganic Materials 37, 871–885 (2001).\n" +
        "    (DOI: 10.1023/A:1011625728803)\n" + 
        "bondi: radii from J. Phys. Chem. 1964, 68, 3, 441–451 (DOI: 10.1021/j100785a001)\n" +
        "Default: bondi"
    )
    
    sterimol_parser.add_argument(
        "-l", "--old-l",
        action="store_true",
        required=False,
        dest="old_L",
        help="approximate FORTRAN Sterimol method for determining L\n"
        "This is 0.4 + the ideal bond length for a target-H bond\n"
        "Default: L value is from VDW radii of target atom to outer\n"
        "VDW radii of atoms projected onto L-axis"
    )

    sterimol_parser.add_argument(
        "-al", "--at-L",
        default=[None],
        dest="L_value",
        type=lambda x: [float(v) for v in x.split(",")],
        help="get widths at specific L values (comma-separated)\n"
        "Default: use the entire ligand",
    )

    sterimol_parser.add_argument(
        "-v", "--vector",
        action="store_true",
        required=False,
        dest="vector",
        help="print Chimera/ChimeraX bild file for vectors instead of parameter values"
    )
    
    sterimol_parser.add_argument(
        "-o", "--output",
        type=str,
        default=False,
        required=False,
        metavar="output destination",
        dest="outfile",
        help="output destination\n" +
        "Default: stdout"
    )
    
    args = sterimol_parser.parse_args(args=argv)
    
    s = ""
    if not args.vector:
        s += "B1\tB2\tB3\tB4\tB5\tL\tfile\n"
    
    for infile in glob_files(args.infile, parser=sterimol_parser):
        if isinstance(infile, str):
            if args.input_format is not None:
                f = FileReader((infile, args.input_format, infile))
            else:
                f = FileReader(infile)
        else:
            if args.input_format is not None:
                f = FileReader(("from stdin", args.input_format, infile))
            else:
                f = FileReader(("from stdin", "xyz", infile))
    
        geom = Geometry(f)
        target = args.targets
        avoid = args.avoid
        end = geom.find(avoid)[0]
        frag = geom.get_fragment(target, stop=end)
        sub = Substituent(frag, end=end, detect=False)
        for val in args.L_value:
            data = sub.sterimol(
                return_vector=args.vector,
                radii=args.radii,
                old_L=args.old_L,
                at_L=val,
            )
            if args.vector:
                for key, color in zip(
                        ["B1", "B2", "B3", "B4", "B5", "L"],
                        ["black", "green", "purple", "orange", "red", "blue"]
                ):
                    start, end = data[key]
                    s += ".color %s\n" % color
                    s += ".note Sterimol %s\n" % key
                    s += ".arrow %6.3f %6.3f %6.3f   %6.3f %6.3f %6.3f\n" % (*start, *end)
            else:
                s += "%.2f\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\t%s\n" % (
                    data["B1"],
                    data["B2"],
                    data["B3"],
                    data["B4"],
                    data["B5"],
                    data["L"],
                    infile,
                )
    
    if not args.outfile:
        print(s)
    else:
        with open(args.outfile, "w") as f:
            f.write(s)

if __name__ == "__main__":
    main(None)



#!/usr/bin/env python3

import argparse
import sys

from AaronTools.fileIO import FileReader
from AaronTools.utils.utils import get_filename, glob_files

from matplotlib import rcParams
import matplotlib.pyplot as plt

import numpy as np

rcParams["savefig.dpi"] = 300


peak_types = ["pseudo-voigt", "gaussian", "lorentzian", "delta"]
plot_types = ["transmittance", "transmittance-velocity", "uv-vis", "uv-vis-velocity", "ecd", "ecd-velocity"]


def peak_type(x):
    out = [y for y in peak_types if y.startswith(x)]
    if out:
        return out[0]
    raise TypeError(
        "peak type must be one of: %s" % ", ".join(
            peak_types
        )
    )

def plot_type(x):
    out = [y for y in plot_types if y.startswith(x)]
    if out:
        return out[0]
    raise TypeError(
        "plot type must be one of: %s" % ", ".join(
            plot_types
        )
    )


uvvis_parser = argparse.ArgumentParser(
    description="plot UV/vis spectrum",
    formatter_class=argparse.RawTextHelpFormatter
)

uvvis_parser.add_argument(
    "infiles", metavar="files",
    type=str,
    nargs="+",
    help="TD-DFT or EOM job output file(s)"
)

uvvis_parser.add_argument(
    "-o", "--output",
    type=str,
    default=None,
    dest="outfile",
    help="output destination\nDefault: show plot",
)

uvvis_parser.add_argument(
    "-t", "--plot-type",
    type=plot_type,
    choices=plot_types,
    default="uv-vis-velocity",
    dest="plot_type",
    help="type of plot\nDefault: uv-vis-velocity (absorbance)",
)

uvvis_parser.add_argument(
    "-u", "--transient",
    action="store_true",
    dest="transient",
    help="use transient excitation data",
)

uvvis_parser.add_argument(
    "-ev", "--electron-volt",
    action="store_true",
    default=False,
    dest="ev_unit",
    help="use eV on x axis instead of nm",
)


peak_options = uvvis_parser.add_argument_group("peak options")
peak_options.add_argument(
    "-p", "--peak-type",
    type=peak_type,
    choices=peak_types,
    default="gaussian",
    dest="peak_type",
    help="function for peaks\nDefault: gaussian",
)

peak_options.add_argument(
    "-m", "--voigt-mixing",
    type=float,
    default=0.5,
    dest="voigt_mixing",
    help="fraction of pseudo-Voigt that is Gaussian\nDefault: 0.5",
)

peak_options.add_argument(
    "-fwhm", "--full-width-half-max",
    type=float,
    default=0.5,
    dest="fwhm",
    help="full width at half max. of peaks\nDefault: 0.5 eV",
)

uvvis_parser.add_argument(
    "-s", "--point-spacing",
    default=None,
    type=float,
    dest="point_spacing",
    help="spacing between each x value\n"
    "Default: a non-uniform spacing that is more dense near peaks",
)


scale_options = uvvis_parser.add_argument_group("scale energies (in eV)")
scale_options.add_argument(
    "-ss", "--scalar-shift",
    type=float,
    default=0.0,
    dest="scalar_scale",
    help="subtract scalar shift from each excitation\n"
    "Default: 0 (no shift)",
)

scale_options.add_argument(
    "-l", "--linear-scale",
    type=float,
    default=0.0,
    dest="linear_scale",
    help="subtract linear_scale * energy from each excitation\n"
    "Default: 0 (no scaling)",
)

scale_options.add_argument(
    "-q", "--quadratic-scale",
    type=float,
    default=0.0,
    dest="quadratic_scale",
    help="subtract quadratic_scale * energy^2 from each excitation\n"
    "Default: 0 (no scaling)",
)

center_centric = uvvis_parser.add_argument_group("x-centered interruptions")
center_centric.add_argument(
    "-sc", "--section-centers",
    type=lambda x: [float(v) for v in x.split(",")],
    dest="centers",
    default=None,
    help="split plot into sections with a section centered on each of the specified values\n"
    "values should be separated by commas"
)

center_centric.add_argument(
    "-sw", "--section-widths",
    type=lambda x: [float(v) for v in x.split(",")],
    dest="widths",
    default=None,
    help="width of each section specified by -c/--centers\n"
    "should be separated by commas, with one for each section"
)

minmax_centric = uvvis_parser.add_argument_group("x-range interruptions")
minmax_centric.add_argument(
    "-r", "--ranges",
    type=lambda x: [[float(v) for v in r.split("-")] for r in x.split(",")],
    dest="ranges",
    default=None,
    help="split plot into sections (e.g. 200-350,400-650)"
)

uvvis_parser.add_argument(
    "-fw", "--figure-width",
    type=float,
    dest="fig_width",
    help="width of figure in inches"
)

uvvis_parser.add_argument(
    "-fh", "--figure-height",
    type=float,
    dest="fig_height",
    help="height of figure in inches"
)

uvvis_parser.add_argument(
    "-csv", "--experimental-csv",
    type=str,
    nargs="+",
    dest="exp_data",
    help="CSV file containing observed spectrum data, which will be plotted on top\n"
    "frequency job files should not come directly after this flag"
)

uvvis_parser.add_argument(
    "-rx", "--rotate-x-ticks",
    action="store_true",
    dest="rotate_x_ticks",
    default=False,
    help="rotate x-axis tick labels by 45 degrees"
)

args = uvvis_parser.parse_args()


if bool(args.centers) != bool(args.widths):
    sys.stderr.write(
        "both -sw/--section-widths and -sc/--section-centers must be specified"
    )
    sys.exit(2)

if args.ranges and bool(args.ranges) == bool(args.widths):
    sys.stderr.write(
        "cannot use -r/--ranges with -sw/--section-widths"
    )
    sys.exit(2)
   
centers = args.centers
widths = args.widths
if args.ranges:
    centers = []
    widths = []
    for (xmin, xmax) in args.ranges:
        centers.append((xmin + xmax) / 2)
        widths.append(abs(xmax - xmin))

units = "nm"
if args.ev_unit:
    units = "eV"

exp_data = None
if args.exp_data:
    exp_data = []
    for f in args.exp_data:
        data = np.loadtxt(f, delimiter=",")
        
        for i in range(1, data.shape[1]):
            exp_data.append((data[:,0], data[:,i], None))

for f in glob_files(args.infiles, parser=uvvis_parser):
    fr = FileReader(f, just_geom=False)

    uv_vis = fr.other["uv_vis"]

    fig = plt.gcf()
    fig.clear()

    uv_vis.plot_uv_vis(
        fig,
        centers=centers,
        widths=widths,
        plot_type=args.plot_type,
        peak_type=args.peak_type,
        fwhm=args.fwhm,
        point_spacing=args.point_spacing,
        voigt_mixing=args.voigt_mixing,
        scalar_scale=args.scalar_scale,
        linear_scale=args.linear_scale,
        quadratic_scale=args.quadratic_scale,
        exp_data=exp_data,
        units=units,
        rotate_x_ticks=args.rotate_x_ticks,
        transient=args.transient,
    )

    if args.fig_width:
        fig.set_figwidth(args.fig_width)

    if args.fig_height:
        fig.set_figheight(args.fig_height)

    if args.outfile:
        outfile_name = args.outfile.replace("$INFILE", get_filename(f))
        plt.savefig(outfile_name, dpi=300)
    else:
        plt.show()



#!/usr/bin/env python3

import argparse
import os
import sys

import numpy as np
from AaronTools.geometry import Geometry
from AaronTools.substituent import Substituent

libaddsub_parser = argparse.ArgumentParser(
    description="add a substituent to your personal library",
    formatter_class=argparse.RawTextHelpFormatter,
)

libaddsub_parser.add_argument(
    "infile",
    metavar="input file",
    type=str,
    default=None,
    help="a coordinate file",
)

libaddsub_parser.add_argument(
    "-n",
    "--name",
    type=str,
    required=False,
    default=None,
    dest="name",
    help="Name of substituent being added to the library\n" +
    "if no name is given, the substituent will be printed to STDOUT",
)

libaddsub_parser.add_argument(
    "-s",
    "--substituent-atom",
    type=str,
    nargs="+",
    required=True,
    dest="target",
    help="substituent atom connected to the rest of the molecule (1-indexed)",
)

libaddsub_parser.add_argument(
    "-a",
    "--attached-to",
    type=str,
    nargs="+",
    required=True,
    dest="avoid",
    help="atom on the molecule that is connected to the substituent",
)

libaddsub_parser.add_argument(
    "-c",
    "--conf",
    type=int,
    nargs=2,
    required=True,
    metavar=("CONFORMERS", "ANGLE"),
    dest="confangle",
    help="number of conformers and the rotation angle (degrees) used to generate each conformer",
)

args = libaddsub_parser.parse_args()

n_confs, angle = args.confangle

if n_confs < 1:
    raise RuntimeError("conformers cannot be < 1")

geom = Geometry(args.infile)
geom.coord_shift(-geom.COM(args.avoid))
sub = geom.get_fragment(args.target, args.avoid, as_object=True)

target = geom.COM(args.target)
x_axis = np.array([1.0, 0.0, 0.0])
n = np.linalg.norm(target)
vb = target / n
d = np.linalg.norm(vb - x_axis)
theta = np.arccos((d ** 2 - 2) / -2)
vx = np.cross(vb, x_axis)
sub.rotate(vx, theta)

sub.comment = "CF:%i,%i" % (n_confs, angle)

if args.name is None:
    print(sub.write(outfile=False))
else:
    sub_file = os.path.join(
        os.path.dirname(Substituent.AARON_LIBS), args.name + ".xyz"
    )
    if os.path.exists(sub_file):
        overwrite = input(
            "%s already exists.\nWould you like to overwrite it? (yes/NO)\n"
            % sub_file
        )
        if overwrite.lower() not in ["yes", "y"]:
            print("not overwriting")
            sys.exit(0)

    sub.write(outfile=sub_file)
